@@ -5,25 +5,33 @@ import (
 	"context"
 	"database/sql"
 	"encoding/json"
+	"errors"
+	"fmt"
 	"math"
 	"os"
 	"reflect"
 	"regexp"
+	"strconv"
+	"sync"
 	"testing"
 	"time"
 
-	"github.com/Masterminds/semver/v3"
-	"github.com/jackc/pgconn"
-	"github.com/jackc/pgx/v4"
-	"github.com/jackc/pgx/v4/stdlib"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
+
+	"github.com/jackc/pgx/v5"
+	"github.com/jackc/pgx/v5/pgconn"
+	"github.com/jackc/pgx/v5/pgtype"
+	"github.com/jackc/pgx/v5/pgxpool"
+	"github.com/jackc/pgx/v5/stdlib"
+	"github.com/jackc/pgx/v5/tracelog"
 )
 
-func openDB(t testing.TB) *sql.DB {
+func openDB(t testing.TB, opts ...stdlib.OptionOpenDB) *sql.DB {
+	t.Helper()
 	config, err := pgx.ParseConfig(os.Getenv("PGX_TEST_DATABASE"))
 	require.NoError(t, err)
-	return stdlib.OpenDB(*config)
+	return stdlib.OpenDB(*config, opts...)
 }
 
 func closeDB(t testing.TB, db *sql.DB) {
@@ -36,7 +44,7 @@ func skipCockroachDB(t testing.TB, db *sql.DB, msg string) {
 	require.NoError(t, err)
 	defer conn.Close()
 
-	err = conn.Raw(func(driverConn interface{}) error {
+	err = conn.Raw(func(driverConn any) error {
 		conn := driverConn.(*stdlib.Conn).Conn()
 		if conn.PgConn().ParameterStatus("crdb_version") != "" {
 			t.Skip(msg)
@@ -46,77 +54,64 @@ func skipCockroachDB(t testing.TB, db *sql.DB, msg string) {
 	require.NoError(t, err)
 }
 
-func skipPostgreSQLVersion(t testing.TB, db *sql.DB, constraintStr, msg string) {
+func skipPostgreSQLVersionLessThan(t testing.TB, db *sql.DB, minVersion int64) {
 	conn, err := db.Conn(context.Background())
 	require.NoError(t, err)
 	defer conn.Close()
 
-	err = conn.Raw(func(driverConn interface{}) error {
+	err = conn.Raw(func(driverConn any) error {
 		conn := driverConn.(*stdlib.Conn).Conn()
 		serverVersionStr := conn.PgConn().ParameterStatus("server_version")
-		serverVersionStr = regexp.MustCompile(`^[0-9.]+`).FindString(serverVersionStr)
+		serverVersionStr = regexp.MustCompile(`^[0-9]+`).FindString(serverVersionStr)
 		// if not PostgreSQL do nothing
 		if serverVersionStr == "" {
 			return nil
 		}
 
-		serverVersion, err := semver.NewVersion(serverVersionStr)
+		serverVersion, err := strconv.ParseInt(serverVersionStr, 10, 64)
 		if err != nil {
 			return err
 		}
 
-		c, err := semver.NewConstraint(constraintStr)
-		if err != nil {
-			return err
+		if serverVersion < minVersion {
+			t.Skipf("Test requires PostgreSQL v%d+", minVersion)
 		}
 
-		if c.Check(serverVersion) {
-			t.Skip(msg)
-		}
 		return nil
 	})
 	require.NoError(t, err)
 }
 
-func testWithAndWithoutPreferSimpleProtocol(t *testing.T, f func(t *testing.T, db *sql.DB)) {
-	t.Run("SimpleProto",
-		func(t *testing.T) {
-			config, err := pgx.ParseConfig(os.Getenv("PGX_TEST_DATABASE"))
-			require.NoError(t, err)
-
-			config.PreferSimpleProtocol = true
-			db := stdlib.OpenDB(*config)
-			defer func() {
-				err := db.Close()
+func testWithAllQueryExecModes(t *testing.T, f func(t *testing.T, db *sql.DB)) {
+	for _, mode := range []pgx.QueryExecMode{
+		pgx.QueryExecModeCacheStatement,
+		pgx.QueryExecModeCacheDescribe,
+		pgx.QueryExecModeDescribeExec,
+		pgx.QueryExecModeExec,
+		pgx.QueryExecModeSimpleProtocol,
+	} {
+		t.Run(mode.String(),
+			func(t *testing.T) {
+				config, err := pgx.ParseConfig(os.Getenv("PGX_TEST_DATABASE"))
 				require.NoError(t, err)
-			}()
 
-			f(t, db)
+				config.DefaultQueryExecMode = mode
+				db := stdlib.OpenDB(*config)
+				defer func() {
+					err := db.Close()
+					require.NoError(t, err)
+				}()
 
-			ensureDBValid(t, db)
-		},
-	)
+				f(t, db)
 
-	t.Run("DefaultProto",
-		func(t *testing.T) {
-			config, err := pgx.ParseConfig(os.Getenv("PGX_TEST_DATABASE"))
-			require.NoError(t, err)
-
-			db := stdlib.OpenDB(*config)
-			defer func() {
-				err := db.Close()
-				require.NoError(t, err)
-			}()
-
-			f(t, db)
-
-			ensureDBValid(t, db)
-		},
-	)
+				ensureDBValid(t, db)
+			},
+		)
+	}
 }
 
 // Do a simple query to ensure the DB is still usable. This is of less use in stdlib as the connection pool should
-// cover an broken connections.
+// cover broken connections.
 func ensureDBValid(t testing.TB, db *sql.DB) {
 	var sum, rowCount int32
 
@@ -161,11 +156,10 @@ func TestSQLOpen(t *testing.T) {
 		driverName string
 	}{
 		{driverName: "pgx"},
-		{driverName: "pgx/v4"},
+		{driverName: "pgx/v5"},
 	}
 
 	for _, tt := range tests {
-		tt := tt
 
 		t.Run(tt.driverName, func(t *testing.T) {
 			db, err := sql.Open(tt.driverName, os.Getenv("PGX_TEST_DATABASE"))
@@ -175,6 +169,17 @@ func TestSQLOpen(t *testing.T) {
 	}
 }
 
+func TestSQLOpenFromPool(t *testing.T) {
+	pool, err := pgxpool.New(context.Background(), os.Getenv("PGX_TEST_DATABASE"))
+	require.NoError(t, err)
+	t.Cleanup(pool.Close)
+
+	db := stdlib.OpenDBFromPool(pool)
+	ensureDBValid(t, db)
+
+	db.Close()
+}
+
 func TestNormalLifeCycle(t *testing.T) {
 	db := openDB(t)
 	defer closeDB(t, db)
@@ -284,7 +289,7 @@ func TestQueryCloseRowsEarly(t *testing.T) {
 }
 
 func TestConnExec(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		_, err := db.Exec("create temporary table t(a varchar not null)")
 		require.NoError(t, err)
 
@@ -298,7 +303,7 @@ func TestConnExec(t *testing.T) {
 }
 
 func TestConnQuery(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		skipCockroachDB(t, db, "Server issues incorrect ParameterDescription (https://github.com/cockroachdb/cockroach/issues/60907)")
 
 		rows, err := db.Query("select 'foo', n from generate_series($1::int, $2::int) n", int32(1), int32(10))
@@ -328,9 +333,105 @@ func TestConnQuery(t *testing.T) {
 	})
 }
 
+func TestConnConcurrency(t *testing.T) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		_, err := db.Exec("create table t (id integer primary key, str text, dur_str interval)")
+		require.NoError(t, err)
+
+		defer func() {
+			_, err := db.Exec("drop table t")
+			require.NoError(t, err)
+		}()
+
+		var wg sync.WaitGroup
+
+		concurrency := 50
+		errChan := make(chan error, concurrency)
+
+		for i := 1; i <= concurrency; i++ {
+			wg.Add(1)
+
+			go func(idx int) {
+				defer wg.Done()
+
+				ctx, cancel := context.WithTimeout(context.Background(), 120*time.Second)
+				defer cancel()
+
+				str := strconv.Itoa(idx)
+				duration := time.Duration(idx) * time.Second
+				_, err := db.ExecContext(ctx, "insert into t values($1)", idx)
+				if err != nil {
+					errChan <- fmt.Errorf("insert failed: %d %w", idx, err)
+					return
+				}
+				_, err = db.ExecContext(ctx, "update t set str = $1 where id = $2", str, idx)
+				if err != nil {
+					errChan <- fmt.Errorf("update 1 failed: %d %w", idx, err)
+					return
+				}
+				_, err = db.ExecContext(ctx, "update t set dur_str = $1 where id = $2", duration, idx)
+				if err != nil {
+					errChan <- fmt.Errorf("update 2 failed: %d %w", idx, err)
+					return
+				}
+
+				errChan <- nil
+			}(i)
+		}
+		wg.Wait()
+		for i := 1; i <= concurrency; i++ {
+			err := <-errChan
+			require.NoError(t, err)
+		}
+
+		for i := 1; i <= concurrency; i++ {
+			wg.Add(1)
+
+			go func(idx int) {
+				defer wg.Done()
+
+				ctx, cancel := context.WithTimeout(context.Background(), 120*time.Second)
+				defer cancel()
+
+				var id int
+				var str string
+				var duration pgtype.Interval
+				err := db.QueryRowContext(ctx, "select id,str,dur_str from t where id = $1", idx).Scan(&id, &str, &duration)
+				if err != nil {
+					errChan <- fmt.Errorf("select failed: %d %w", idx, err)
+					return
+				}
+				if id != idx {
+					errChan <- fmt.Errorf("id mismatch: %d %d", idx, id)
+					return
+				}
+				if str != strconv.Itoa(idx) {
+					errChan <- fmt.Errorf("str mismatch: %d %s", idx, str)
+					return
+				}
+				expectedDuration := pgtype.Interval{
+					Microseconds: int64(idx) * time.Second.Microseconds(),
+					Valid:        true,
+				}
+				if duration != expectedDuration {
+					errChan <- fmt.Errorf("duration mismatch: %d %v", idx, duration)
+					return
+				}
+
+				errChan <- nil
+			}(i)
+		}
+		wg.Wait()
+		for i := 1; i <= concurrency; i++ {
+			err := <-errChan
+			require.NoError(t, err)
+		}
+	})
+}
+
 // https://github.com/jackc/pgx/issues/781
 func TestConnQueryDifferentScanPlansIssue781(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		var s string
 		var b bool
 
@@ -345,7 +446,7 @@ func TestConnQueryDifferentScanPlansIssue781(t *testing.T) {
 }
 
 func TestConnQueryNull(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		rows, err := db.Query("select $1::int", nil)
 		require.NoError(t, err)
 
@@ -370,7 +471,7 @@ func TestConnQueryNull(t *testing.T) {
 }
 
 func TestConnQueryRowByteSlice(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		expected := []byte{222, 173, 190, 239}
 		var actual []byte
 
@@ -381,15 +482,15 @@ func TestConnQueryRowByteSlice(t *testing.T) {
 }
 
 func TestConnQueryFailure(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		_, err := db.Query("select 'foo")
 		require.Error(t, err)
-		require.IsType(t, new(pgconn.PgError), err)
+		require.ErrorAs(t, err, new(*pgconn.PgError))
 	})
 }
 
 func TestConnSimpleSlicePassThrough(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		skipCockroachDB(t, db, "Server does not support cardinality function")
 
 		var n int64
@@ -399,10 +500,58 @@ func TestConnSimpleSlicePassThrough(t *testing.T) {
 	})
 }
 
+func TestConnQueryScanGoArray(t *testing.T) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		m := pgtype.NewMap()
+
+		var a []int64
+		err := db.QueryRow("select '{1,2,3}'::bigint[]").Scan(m.SQLScanner(&a))
+		require.NoError(t, err)
+		assert.Equal(t, []int64{1, 2, 3}, a)
+	})
+}
+
+func TestConnQueryScanArray(t *testing.T) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		m := pgtype.NewMap()
+
+		var a pgtype.Array[int64]
+		err := db.QueryRow("select '{1,2,3}'::bigint[]").Scan(m.SQLScanner(&a))
+		require.NoError(t, err)
+		assert.Equal(t, pgtype.Array[int64]{Elements: []int64{1, 2, 3}, Dims: []pgtype.ArrayDimension{{Length: 3, LowerBound: 1}}, Valid: true}, a)
+
+		err = db.QueryRow("select null::bigint[]").Scan(m.SQLScanner(&a))
+		require.NoError(t, err)
+		assert.Equal(t, pgtype.Array[int64]{Elements: nil, Dims: nil, Valid: false}, a)
+	})
+}
+
+func TestConnQueryScanRange(t *testing.T) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		skipCockroachDB(t, db, "Server does not support int4range")
+
+		m := pgtype.NewMap()
+
+		var r pgtype.Range[pgtype.Int4]
+		err := db.QueryRow("select int4range(1, 5)").Scan(m.SQLScanner(&r))
+		require.NoError(t, err)
+		assert.Equal(
+			t,
+			pgtype.Range[pgtype.Int4]{
+				Lower:     pgtype.Int4{Int32: 1, Valid: true},
+				Upper:     pgtype.Int4{Int32: 5, Valid: true},
+				LowerType: pgtype.Inclusive,
+				UpperType: pgtype.Exclusive,
+				Valid:     true,
+			},
+			r)
+	})
+}
+
 // Test type that pgx would handle natively in binary, but since it is not a
 // database/sql native type should be passed through as a string
 func TestConnQueryRowPgxBinary(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		sql := "select $1::int4[]"
 		expected := "{1,2,3}"
 		var actual string
@@ -414,7 +563,7 @@ func TestConnQueryRowPgxBinary(t *testing.T) {
 }
 
 func TestConnQueryRowUnknownType(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		skipCockroachDB(t, db, "Server does not support point type")
 
 		sql := "select $1::point"
@@ -428,7 +577,7 @@ func TestConnQueryRowUnknownType(t *testing.T) {
 }
 
 func TestConnQueryJSONIntoByteSlice(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		_, err := db.Exec(`
 		create temporary table docs(
 			body json not null
@@ -447,7 +596,7 @@ func TestConnQueryJSONIntoByteSlice(t *testing.T) {
 			t.Errorf("Unexpected failure: %v (sql -> %v)", err, sql)
 		}
 
-		if bytes.Compare(actual, expected) != 0 {
+		if !bytes.Equal(actual, expected) {
 			t.Errorf(`Expected "%v", got "%v" (sql -> %v)`, string(expected), string(actual), sql)
 		}
 
@@ -479,7 +628,7 @@ func TestConnExecInsertByteSliceIntoJSON(t *testing.T) {
 	err = db.QueryRow(`select body from docs`).Scan(&actual)
 	require.NoError(t, err)
 
-	if bytes.Compare(actual, expected) != 0 {
+	if !bytes.Equal(actual, expected) {
 		t.Errorf(`Expected "%v", got "%v"`, string(expected), string(actual))
 	}
 
@@ -488,7 +637,7 @@ func TestConnExecInsertByteSliceIntoJSON(t *testing.T) {
 }
 
 func TestTransactionLifeCycle(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		_, err := db.Exec("create temporary table t(a varchar not null)")
 		require.NoError(t, err)
 
@@ -522,7 +671,7 @@ func TestTransactionLifeCycle(t *testing.T) {
 }
 
 func TestConnBeginTxIsolation(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		skipCockroachDB(t, db, "Server always uses serializable isolation level")
 
 		var defaultIsoLevel string
@@ -578,7 +727,7 @@ func TestConnBeginTxIsolation(t *testing.T) {
 }
 
 func TestConnBeginTxReadOnly(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		tx, err := db.BeginTx(context.Background(), &sql.TxOptions{ReadOnly: true})
 		require.NoError(t, err)
 		defer tx.Rollback()
@@ -596,7 +745,7 @@ func TestConnBeginTxReadOnly(t *testing.T) {
 }
 
 func TestBeginTxContextCancel(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		_, err := db.Exec("drop table if exists t")
 		require.NoError(t, err)
 
@@ -617,55 +766,20 @@ func TestBeginTxContextCancel(t *testing.T) {
 
 		var n int
 		err = db.QueryRow("select count(*) from t").Scan(&n)
-		if pgErr, ok := err.(*pgconn.PgError); !ok || pgErr.Code != "42P01" {
+		var pgErr *pgconn.PgError
+		if !errors.As(err, &pgErr) || pgErr.Code != "42P01" {
 			t.Fatalf(`err => %v, want PgError{Code: "42P01"}`, err)
 		}
 	})
 }
 
-func TestAcquireConn(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
-		var conns []*pgx.Conn
-
-		for i := 1; i < 6; i++ {
-			conn, err := stdlib.AcquireConn(db)
-			if err != nil {
-				t.Errorf("%d. AcquireConn failed: %v", i, err)
-				continue
-			}
-
-			var n int32
-			err = conn.QueryRow(context.Background(), "select 1").Scan(&n)
-			if err != nil {
-				t.Errorf("%d. QueryRow failed: %v", i, err)
-			}
-			if n != 1 {
-				t.Errorf("%d. n => %d, want %d", i, n, 1)
-			}
-
-			stats := db.Stats()
-			if stats.OpenConnections != i {
-				t.Errorf("%d. stats.OpenConnections => %d, want %d", i, stats.OpenConnections, i)
-			}
-
-			conns = append(conns, conn)
-		}
-
-		for i, conn := range conns {
-			if err := stdlib.ReleaseConn(db, conn); err != nil {
-				t.Errorf("%d. stdlib.ReleaseConn failed: %v", i, err)
-			}
-		}
-	})
-}
-
 func TestConnRaw(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		conn, err := db.Conn(context.Background())
 		require.NoError(t, err)
 
 		var n int
-		err = conn.Raw(func(driverConn interface{}) error {
+		err = conn.Raw(func(driverConn any) error {
 			conn := driverConn.(*stdlib.Conn).Conn()
 			return conn.QueryRow(context.Background(), "select 42").Scan(&n)
 		})
@@ -674,47 +788,15 @@ func TestConnRaw(t *testing.T) {
 	})
 }
 
-// https://github.com/jackc/pgx/issues/673
-func TestReleaseConnWithTxInProgress(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
-		skipCockroachDB(t, db, "Server does not support backend PID")
-
-		c1, err := stdlib.AcquireConn(db)
-		require.NoError(t, err)
-
-		_, err = c1.Exec(context.Background(), "begin")
-		require.NoError(t, err)
-
-		c1PID := c1.PgConn().PID()
-
-		err = stdlib.ReleaseConn(db, c1)
-		require.NoError(t, err)
-
-		c2, err := stdlib.AcquireConn(db)
-		require.NoError(t, err)
-
-		c2PID := c2.PgConn().PID()
-
-		err = stdlib.ReleaseConn(db, c2)
-		require.NoError(t, err)
-
-		require.NotEqual(t, c1PID, c2PID)
-
-		// Releasing a conn with a tx in progress should close the connection
-		stats := db.Stats()
-		require.Equal(t, 1, stats.OpenConnections)
-	})
-}
-
 func TestConnPingContextSuccess(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		err := db.PingContext(context.Background())
 		require.NoError(t, err)
 	})
 }
 
 func TestConnPrepareContextSuccess(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		stmt, err := db.PrepareContext(context.Background(), "select now()")
 		require.NoError(t, err)
 		err = stmt.Close()
@@ -722,32 +804,43 @@ func TestConnPrepareContextSuccess(t *testing.T) {
 	})
 }
 
+// https://github.com/jackc/pgx/issues/1753#issuecomment-1746033281
+// https://github.com/jackc/pgx/issues/1754#issuecomment-1752004634
+func TestConnMultiplePrepareAndDeallocate(t *testing.T) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		skipCockroachDB(t, db, "Server does not support pg_prepared_statements")
+
+		sql := "select 42"
+		stmt1, err := db.PrepareContext(context.Background(), sql)
+		require.NoError(t, err)
+		stmt2, err := db.PrepareContext(context.Background(), sql)
+		require.NoError(t, err)
+		err = stmt1.Close()
+		require.NoError(t, err)
+
+		var preparedStmtCount int64
+		err = db.QueryRowContext(context.Background(), "select count(*) from pg_prepared_statements where statement = $1", sql).Scan(&preparedStmtCount)
+		require.NoError(t, err)
+		require.EqualValues(t, 1, preparedStmtCount)
+
+		err = stmt2.Close() // err isn't as useful as it should be as database/sql will ignore errors from Deallocate.
+		require.NoError(t, err)
+
+		err = db.QueryRowContext(context.Background(), "select count(*) from pg_prepared_statements where statement = $1", sql).Scan(&preparedStmtCount)
+		require.NoError(t, err)
+		require.EqualValues(t, 0, preparedStmtCount)
+	})
+}
+
 func TestConnExecContextSuccess(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		_, err := db.ExecContext(context.Background(), "create temporary table exec_context_test(id serial primary key)")
 		require.NoError(t, err)
 	})
 }
 
-func TestConnExecContextFailureRetry(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
-		// We get a connection, immediately close it, and then get it back;
-		// DB.Conn along with Conn.ResetSession does the retry for us.
-		{
-			conn, err := stdlib.AcquireConn(db)
-			require.NoError(t, err)
-			conn.Close(context.Background())
-			stdlib.ReleaseConn(db, conn)
-		}
-		conn, err := db.Conn(context.Background())
-		require.NoError(t, err)
-		_, err = conn.ExecContext(context.Background(), "select 1")
-		require.NoError(t, err)
-	})
-}
-
 func TestConnQueryContextSuccess(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		rows, err := db.QueryContext(context.Background(), "select * from generate_series(1,10) n")
 		require.NoError(t, err)
 
@@ -760,26 +853,8 @@ func TestConnQueryContextSuccess(t *testing.T) {
 	})
 }
 
-func TestConnQueryContextFailureRetry(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
-		// We get a connection, immediately close it, and then get it back;
-		// DB.Conn along with Conn.ResetSession does the retry for us.
-		{
-			conn, err := stdlib.AcquireConn(db)
-			require.NoError(t, err)
-			conn.Close(context.Background())
-			stdlib.ReleaseConn(db, conn)
-		}
-		conn, err := db.Conn(context.Background())
-		require.NoError(t, err)
-
-		_, err = conn.QueryContext(context.Background(), "select 1")
-		require.NoError(t, err)
-	})
-}
-
 func TestRowsColumnTypeDatabaseTypeName(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		rows, err := db.Query("select 42::bigint")
 		require.NoError(t, err)
 
@@ -863,7 +938,7 @@ func TestStmtQueryContextSuccess(t *testing.T) {
 }
 
 func TestRowsColumnTypes(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		columnTypesTests := []struct {
 			Name     string
 			TypeName string
@@ -1001,7 +1076,7 @@ func TestRowsColumnTypes(t *testing.T) {
 }
 
 func TestQueryLifeCycle(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		skipCockroachDB(t, db, "Server issues incorrect ParameterDescription (https://github.com/cockroachdb/cockroach/issues/60907)")
 
 		rows, err := db.Query("SELECT 'foo', n FROM generate_series($1::int, $2::int) n WHERE 3 = $3", 1, 10, 3)
@@ -1050,7 +1125,7 @@ func TestQueryLifeCycle(t *testing.T) {
 
 // https://github.com/jackc/pgx/issues/409
 func TestScanJSONIntoJSONRawMessage(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
 		var msg json.RawMessage
 
 		err := db.QueryRow("select '{}'::json").Scan(&msg)
@@ -1060,16 +1135,16 @@ func TestScanJSONIntoJSONRawMessage(t *testing.T) {
 }
 
 type testLog struct {
-	lvl  pgx.LogLevel
+	lvl  tracelog.LogLevel
 	msg  string
-	data map[string]interface{}
+	data map[string]any
 }
 
 type testLogger struct {
 	logs []testLog
 }
 
-func (l *testLogger) Log(ctx context.Context, lvl pgx.LogLevel, msg string, data map[string]interface{}) {
+func (l *testLogger) Log(ctx context.Context, lvl tracelog.LogLevel, msg string, data map[string]any) {
 	l.logs = append(l.logs, testLog{lvl: lvl, msg: msg, data: data})
 }
 
@@ -1078,7 +1153,7 @@ func TestRegisterConnConfig(t *testing.T) {
 	require.NoError(t, err)
 
 	logger := &testLogger{}
-	connConfig.Logger = logger
+	connConfig.Tracer = &tracelog.TraceLog{Logger: logger, LogLevel: tracelog.LogLevelInfo}
 
 	// Issue 947: Register and unregister a ConnConfig and ensure that the
 	// returned connection string is not reused.
@@ -1105,8 +1180,8 @@ func TestRegisterConnConfig(t *testing.T) {
 
 // https://github.com/jackc/pgx/issues/958
 func TestConnQueryRowConstraintErrors(t *testing.T) {
-	testWithAndWithoutPreferSimpleProtocol(t, func(t *testing.T, db *sql.DB) {
-		skipPostgreSQLVersion(t, db, "< 11", "Test requires PG 11+")
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		skipPostgreSQLVersionLessThan(t, db, 11)
 		skipCockroachDB(t, db, "Server does not support deferred constraint (https://github.com/cockroachdb/cockroach/issues/31632)")
 
 		_, err := db.Exec(`create temporary table defer_test (
@@ -1184,13 +1259,13 @@ func TestRandomizeHostOrderFunc(t *testing.T) {
 
 	// Test that at some point we connect to all 3 hosts
 	hostsNotSeenYet := map[string]struct{}{
-		"host1": struct{}{},
-		"host2": struct{}{},
-		"host3": struct{}{},
+		"host1": {},
+		"host2": {},
+		"host3": {},
 	}
 
 	// If we don't succeed within this many iterations, something is certainly wrong
-	for i := 0; i < 100000; i++ {
+	for range 100_000 {
 		connCopy := *config
 		stdlib.RandomizeHostOrderFunc(context.Background(), &connCopy)
 
@@ -1238,3 +1313,159 @@ func TestResetSessionHookCalled(t *testing.T) {
 
 	require.True(t, mockCalled)
 }
+
+func TestCheckIdleConn(t *testing.T) {
+	controllerConn, err := sql.Open("pgx", os.Getenv("PGX_TEST_DATABASE"))
+	require.NoError(t, err)
+	defer closeDB(t, controllerConn)
+
+	skipCockroachDB(t, controllerConn, "Server does not support pg_terminate_backend() (https://github.com/cockroachdb/cockroach/issues/35897)")
+
+	db, err := sql.Open("pgx", os.Getenv("PGX_TEST_DATABASE"))
+	require.NoError(t, err)
+	defer closeDB(t, db)
+
+	var conns []*sql.Conn
+	for range 3 {
+		c, err := db.Conn(context.Background())
+		require.NoError(t, err)
+		conns = append(conns, c)
+	}
+
+	require.EqualValues(t, 3, db.Stats().OpenConnections)
+
+	var pids []uint32
+	for _, c := range conns {
+		err := c.Raw(func(driverConn any) error {
+			pids = append(pids, driverConn.(*stdlib.Conn).Conn().PgConn().PID())
+			return nil
+		})
+		require.NoError(t, err)
+		err = c.Close()
+		require.NoError(t, err)
+	}
+
+	// The database/sql connection pool seems to automatically close idle connections to only keep 2 alive.
+	// require.EqualValues(t, 3, db.Stats().OpenConnections)
+
+	_, err = controllerConn.ExecContext(context.Background(), `select pg_terminate_backend(n) from unnest($1::int[]) n`, pids)
+	require.NoError(t, err)
+
+	// All conns are dead they don't know it and neither does the pool. But because of database/sql automatically closing
+	// idle connections we can't be sure how many we should have. require.EqualValues(t, 3, db.Stats().OpenConnections)
+
+	// Wait long enough so the pool will realize it needs to check the connections.
+	time.Sleep(time.Second)
+
+	// Pool should try all existing connections and find them dead, then create a new connection which should successfully ping.
+	err = db.PingContext(context.Background())
+	require.NoError(t, err)
+
+	// The original 3 conns should have been terminated and the a new conn established for the ping.
+	require.EqualValues(t, 1, db.Stats().OpenConnections)
+	c, err := db.Conn(context.Background())
+	require.NoError(t, err)
+
+	var cPID uint32
+	err = c.Raw(func(driverConn any) error {
+		cPID = driverConn.(*stdlib.Conn).Conn().PgConn().PID()
+		return nil
+	})
+	require.NoError(t, err)
+	err = c.Close()
+	require.NoError(t, err)
+
+	require.NotContains(t, pids, cPID)
+}
+
+func TestOptionShouldPing_HookCalledOnReuse(t *testing.T) {
+	hookCalled := false
+
+	db := openDB(t,
+		stdlib.OptionShouldPing(func(context.Context, stdlib.ShouldPingParams) bool {
+			hookCalled = true
+			// Return false to avoid relying on actual ping behavior.
+			return false
+		}),
+	)
+	defer closeDB(t, db)
+
+	// Ensure reuse (so ResetSession runs)
+	db.SetMaxOpenConns(1)
+	db.SetMaxIdleConns(1)
+
+	// Establish the connection
+	require.NoError(t, db.Ping())
+
+	// Reuse the connection -> should trigger ResetSession -> ShouldPing
+	_, err := db.Exec("select 1")
+	require.NoError(t, err)
+
+	require.True(t, hookCalled, "hook should be called on reuse")
+}
+
+func TestRowsColumnTypeLength(t *testing.T) {
+	testWithAllQueryExecModes(t, func(t *testing.T, db *sql.DB) {
+		skipCockroachDB(t, db, "Server does not support type")
+
+		columnTypeLengthTests := []struct {
+			Len int64
+			OK  bool
+		}{
+			{
+				math.MaxInt64,
+				true,
+			},
+			{
+				math.MaxInt64,
+				true,
+			},
+			{
+				255,
+				true,
+			},
+			{
+				10,
+				true,
+			},
+			{
+				50,
+				true,
+			},
+			{
+				0,
+				false,
+			},
+		}
+
+		_, err := db.Exec(`CREATE TEMPORARY TABLE temp_column_type_length (
+						   text_column TEXT,
+						   bytea_column BYTEA,
+						   varchar_column VARCHAR(255),
+						   bpcharA_column BPCHAR(10)[],
+						   varbit_column VARBIT(50),
+						   int_column INT
+					);`)
+		require.NoError(t, err)
+
+		rows, err := db.Query("SELECT * FROM temp_column_type_length")
+		require.NoError(t, err)
+
+		columns, err := rows.ColumnTypes()
+		require.NoError(t, err)
+		assert.Len(t, columns, 6)
+
+		for i, tt := range columnTypeLengthTests {
+			c := columns[i]
+
+			l, ok := c.Length()
+			if l != tt.Len {
+				t.Errorf("(%d) got: %d, want: %d", i, l, tt.Len)
+			}
+			if ok != tt.OK {
+				t.Errorf("(%d) got: %t, want: %t", i, ok, tt.OK)
+			}
+
+		}
+	})
+}
